#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "colour.h"
#include "atmos.h"
#include "beams.h"
#include "transform.h"
#include "render.h"

#define N_PRECALC_BEAMS 256
/* BILINEAR putpixel on/off, slightly faster when off.. visual quality doesn't change much */
#define BILINEAR 0
/* physically correct spectral integration, a whole lot of uselessness for visualisation */
#define CORRECT_COLOURS 0

float curtain_startpoints[N_TOTAL_STARTPOINTS*N_DIM];
float beams[N_TOTAL_STARTPOINTS*N_MAX_DEFLECTIONS*N_DIM];
float tbeams[N_TOTAL_STARTPOINTS*N_MAX_DEFLECTIONS*N_DIM];
float *precalced_beams_ptrs[N_PRECALC_BEAMS+1];
float precalced_beams[N_PRECALC_BEAMS*N_MAX_DEFLECTIONS*N_DIM];
float *intensity_table;
float x_angle, z_angle;

/* mix two rgb buffers (50/50), used for motion blur */
void mix_crgbuf(unsigned char *dst, unsigned char *src, long len)
{
	long i;
	for(i=0; i<len*3; i++)
		dst[i] = (dst[i]>>1) + (src[i]>>1);
}

/* mix two rgb buffers (75/25), used for motion blur */
void mix_crgbuf_(unsigned char *dst, unsigned char *src, long len)
{
	long i;
	for(i=0; i<len*3; i++)
		dst[i] = dst[i] - (dst[i]>>3) + (src[i]>>3);
}

void render_set_intensity_table(float *table)
{
	intensity_table = table;
}

void render_set_cam_angle(float a)
{
	z_angle = a;
}

void render_set_cam_tilt(float a)
{
	x_angle = a;
}


/* 
 * precalculate some electron deflection paths to speed up visual simulation.
 */
void precalc()
{
	float mag_vec[N_DIM] = {0,0,-1};
	float start_point[N_DIM] = {0,0,0};
	float mag_vec_norm[N_DIM];
	int i;

	/* normalise magnetic field direction */
	float r2=0;
	for (i=0; i<N_DIM; i++)
		r2+=mag_vec[i]*mag_vec[i];
	for (i=0; i<N_DIM; i++)
		mag_vec_norm[i] = mag_vec[i] / sqrt(r2);

	float **beams_pt;
	beams_pt = malloc(sizeof(float **));
	*beams_pt = precalced_beams;
	for (i=0; i<N_PRECALC_BEAMS; i++) {
		//printf("# throw electron down beam nr [%04d]\n", i);
		/* TODO: store start of each precipitated beam in array? */
		//printf("beams=[%08X] beams[0]=[%f]\n", beams, beams[0]);
		//printf("beams_pt=[%08X]\n", beams_pt);
		//printf("# beams_pt=[%08X] *beams_pt=[%08X] **beams_pt=[%f]\n", beams_pt, *beams_pt, **beams_pt);
		precalced_beams_ptrs[i] = *beams_pt;
		precipitate(beams_pt, start_point, mag_vec_norm);
	}
	precalced_beams_ptrs[i] = *beams_pt;
	int total_points = (*beams_pt - beams)/N_DIM;
	//printf("total_points [%d]\n", total_points);
	float *point_pt;
	free(beams_pt);
	//printf("freed.\n");
}

/*----------------------------------------------------------------------------*/
/* main simulation routine: renders to RGB pixel buffer                       */
/*----------------------------------------------------------------------------*/

/*
 * pre:
 * - precalc() should be called prior execution when use_precalc!=0
 */
int render(unsigned char *crgbuf, float *rgbuf, float time_shift, int use_precalc, float amp)
{
	/* values in km, and in 'math' world coordinate system (+z=up) */
	float curtain_start[N_DIM] = {-350,+30,300};
	float curtain_end[N_DIM] = {+350,+30,300};
	/* centre of curtain in 3D, note that this should always match curtain_start,_end ! */
	float centre[N_DIM] = {0,+30,200};
	float mag_vec[N_DIM] = {0,0,-1};
	float mag_vec_norm[N_DIM];
	int i;
	int center[] = {X_CENTER, (Y_CENTER*3)/5};
	/* camera distance */
	float dist=100.0;
	/* translation in world space, replace some stuff now done in cam space.. */
	float trans[] = {0.,+200.,0.};

 	intensity_table = atmos_get_intensity_profile();

	/* normalise magnetic field direction */
	float r2=0;
	for (i=0; i<N_DIM; i++)
		r2+=mag_vec[i]*mag_vec[i];
	for (i=0; i<N_DIM; i++)
		mag_vec_norm[i] = mag_vec[i] / sqrt(r2);

	/* generate curtain startpoint using multiple phase-shifted sheets */
	float *pt = curtain_startpoints, shift=0;
	struct auroral_form form;
	for (i=0; i<N_CURTAINS; i++) {

		shift = CURTAIN_PHASESHIFT*(rand() / ((float) RAND_MAX));

//		generate_curtain_startpoints(pt, N_CURTAIN_STARTPOINTS, curtain_start, curtain_end, mag_vec, shift, 5.5+shift+time_shift, 7.0);
		memcpy((void *) &form, (void *) &rayed_arc_form, sizeof(rayed_arc_form));
		form.wave_start_phase += time_shift+shift;
		form.wave_end_phase += time_shift+shift;
		form.wave_amp = 10.0;
		generate_curtain_startpoints(pt, N_CURTAIN_STARTPOINTS, curtain_start, curtain_end, mag_vec, &form);

		pt+=N_CURTAIN_STARTPOINTS*N_DIM;
		//shift+=CURTAIN_PHASESHIFT;
	}

	#if 0
	/* output start points for visualisation purposes */
	for (i=0; i<N_CURTAIN_STARTPOINTS*N_CURTAINS; i++) {
//		printf("[%03d]: (%f,%f,%f)\n", i, curtain_startpoints[i*N_DIM], curtain_startpoints[i*N_DIM+1], curtain_startpoints[i*N_DIM+2]);
		printf("%f %f\n", curtain_startpoints[i*N_DIM], curtain_startpoints[i*N_DIM+2]);
	}
	#endif

	float **beams_pt;
	beams_pt = malloc(sizeof(float **));
	*beams_pt = beams;
	//for (i=0; i<N_CURTAINS*N_CURTAIN_STARTPOINTS; i+=N_CURTAIN_STARTPOINTS) {
	//for (i=0; i<20; i++) {
	for (i=0; i<N_CURTAINS*N_CURTAIN_STARTPOINTS; i++) {
		//printf("# throw electron down beam nr [%04d]\n", i);
		/* TODO: store start of each precipitated beam in array? */
		//printf("beams=[%08X] beams[0]=[%f]\n", beams, beams[0]);
		//printf("beams_pt=[%08X]\n", beams_pt);
		//printf("# beams_pt=[%08X] *beams_pt=[%08X] **beams_pt=[%f]\n", beams_pt, *beams_pt, **beams_pt);
		if (use_precalc) {
			int ridx = rand() % N_PRECALC_BEAMS;
			float *current_beam, *next_beam;
			current_beam = precalced_beams_ptrs[ridx];
			next_beam = precalced_beams_ptrs[ridx+1];
			long size = next_beam - current_beam;
			/* copy precalced beam with curtain start offset */
			float *dst_pt;
			dst_pt = *beams_pt;
			int defidx;
			//printf("size=[%ld]\n", size);
			for (defidx=0; defidx<size/N_DIM; defidx++) {
				*dst_pt++ = current_beam[defidx*3] + curtain_startpoints[i*3];
				*dst_pt++ = current_beam[defidx*3+1] + curtain_startpoints[i*3+1];
				*dst_pt++ = current_beam[defidx*3+2] + curtain_startpoints[i*3+2];
				//printf("([%f],[%f],[%f])\n", dst_pt[-3], dst_pt[-2], dst_pt[-1]);
			}
			//memcpy(*beams_pt, current_beam, size*sizeof(float));
			*beams_pt += size; 
		} else {
			precipitate(beams_pt, curtain_startpoints+i*3, mag_vec_norm);
		}
	}
	int total_points = (*beams_pt - beams)/N_DIM;
	//printf("total_points [%d]\n", total_points);
	float *point_pt;
	free(beams_pt);
	//printf("freed.\n");

	#if 1

	float mat[N_DIM*N_DIM];
	/* this one used for 'own stubborn' world coordinate system (+y=up) */
	//create_matrix(mat, .48*M_PI, .05*M_PI, 0.f);
	/* this one used for 'math' convention world coordinate system (+z=up) */
	create_matrix(mat, x_angle, 0.f, z_angle);
	//print_matrix(mat);
	transform_points(tbeams, beams, total_points, mat, center, dist, centre);

	#endif

	#if 1

	point_pt = tbeams;
	/* TODO: if emission brightness is proportional to 1/z^2 then is additive behaviour appropriate? */
	for (i=0; i<total_points; i++) {
		float camx,camy,camz,alt,scale;
		double xi, yi, zi, r=0, g=0, b=0;

		camx = point_pt[0];
		camy = point_pt[1];
		camz = point_pt[2];
		/* dot radius proportional to 1/z, so dot area proportional to 1/z^2 */
		scale = dist/(camz+dist);
		scale = scale * scale * scale;
		//printf("scale = [%f]\n", scale);
		alt=beams[i*N_DIM+2];
		point_pt+=N_DIM;

		/* skip if behind camera.. */
		if (camz + dist <= 0.)
			continue;

		//if (alt > 110.)
		//	continue;
		//printf("%f %f\n", camx, camy);
		//printf("i=[%d]\n", i);
		//printf("(x=[%f],y=[%f],alt=[%f])\n", camx, camy, alt);

		#if CORRECT_COLOURS

		struct pixel_spectrum ps;
		ps.i630 = lookup_spec_intensity(atmos_get_profile_630nm(), alt);
		ps.i557 = lookup_spec_intensity(atmos_get_profile_557nm(), alt);
		ps.i427 = lookup_spec_intensity(atmos_get_profile_427nm(), alt);
		//ps.i630 = 0;
		//ps.i557 = 0;
		//ps.i427 = 0;

		/* emission wavelength -> RGB conversion */
		pixelspectrum_to_xyz(&ps, &xi, &yi, &zi);
		xyz_to_rgb(&SMPTEsystem, xi, yi, zi, &r, &g, &b);

		if (r<0.)
			r=0;
		if (g<0.)
			g=0;
		if (b<0.)
			b=0;

		#else 

		r = lookup_spec_intensity(atmos_get_profile_630nm(), alt);
		g = lookup_spec_intensity(atmos_get_profile_557nm(), alt);
		b = lookup_spec_intensity(atmos_get_profile_427nm(), alt);

		#endif

		float intensity = lookup_glob_intensity(intensity_table, alt);
		r *= scale * intensity * amp;
		g *= scale * intensity * amp;
		b *= scale * intensity * amp;

		/* plotting to frame buffer */
		//printf("[%f] km: pos=([%f] [%f]) col=([%f] [%f] [%f])\n", alt, camx, camy, r, g, b);
		/* perform camera space clipping.. */
		if ( (camx < 0.) || (camx >= (float) X_RES) )
			continue;
		if ( (camy < 0.) || (camy >= (float) Y_RES) )
			continue;

		//printf("pixel on screen\n");

		int ix = (int) camx;
		int iy = (int) camy;

		#if BILINEAR

		/* use bi-linear interpolation for the pixel output */
		float xf = camx - (float) ix;
		float yf = camy - (float) iy;

		#if 0
		if (ix == 355 && iy==400)
		{
			printf("camx camy = [%d] [%d], off = [%d]\n", ix, iy, (iy*X_RES + ix)*3 + 0); 
			//printf("red upper left = [%f]\n", 255.* r*(1.-xf)*(1.-yf) );
			printf("red upper left = [%f]\n", rgbuf[(iy*X_RES + ix)*3 + 0] );
		}
		#endif

		//printf("interpolating colours..\n");

		/* output red */
		rgbuf[(iy*X_RES + ix)*3 + 2]       += r*(1.-xf)*(1.-yf);
		rgbuf[(iy*X_RES + ix+1)*3 + 2]     += r*xf*(1.-yf);
		rgbuf[((iy+1)*X_RES + ix)*3 + 2]   += r*(1.-xf)*yf;
		rgbuf[((iy+1)*X_RES + ix+1)*3 + 2] += r*xf*yf;

		/* output green */
		rgbuf[(iy*X_RES + ix)*3 + 1]       += g*(1.-xf)*(1.-yf);
		rgbuf[(iy*X_RES + ix+1)*3 + 1]     += g*xf*(1.-yf);
		rgbuf[((iy+1)*X_RES + ix)*3 + 1]   += g*(1.-xf)*yf;
		rgbuf[((iy+1)*X_RES + ix+1)*3 + 1] += g*xf*yf;

		/* output blue */
		rgbuf[(iy*X_RES + ix)*3 + 0]       += b*(1.-xf)*(1.-yf);
		rgbuf[(iy*X_RES + ix+1)*3 + 0]     += b*xf*(1.-yf);
		rgbuf[((iy+1)*X_RES + ix)*3 + 0]   += b*(1.-xf)*yf;
		rgbuf[((iy+1)*X_RES + ix+1)*3 + 0] += b*xf*yf;

		#else

		rgbuf[(iy*X_RES + ix)*3 + 2]       += r;
		rgbuf[(iy*X_RES + ix)*3 + 1]       += g;
		rgbuf[(iy*X_RES + ix)*3 + 0]       += b;

		#endif
	}

	#endif

	#if 1

	int ix=355, iy=400;
	//printf("camx camy = [%d] [%d], off = [%d]\n", ix, iy, (iy*X_RES + ix)*3 + 0); 
	//printf("red upper left = [%f]\n", 255.* r*(1.-xf)*(1.-yf) );
	//printf("red upper left = [%f]\n", rgbuf[(iy*X_RES + ix)*3 + 0] );

	/* convert float rgb buffer to saturated byte buffer */
	int x,y;
	for (y=0; y<Y_RES; y++) {
		for (x=0; x<X_RES; x++) {
			/* reverse vertical axis.. TODO: where else?! */
			int pix_off = X_RES*(Y_RES-1-y)+x;
			int dst_off = (X_RES*y+x)*3;
			for (i=pix_off*3; i<pix_off*3+3; i++) {
				float f = rgbuf[i];
				if (f > 1.0)
					f = 1.0;
				crgbuf[dst_off] = (unsigned char) (f*255.);
				//if (dst_off==961065)
				//	printf("f=[%f], c=[%d]\n", f, (int) crgbuf[dst_off]);
				dst_off++;
			}
		}
	}

	#endif

	return 0;
}

int render_init()
{
	precalc();
	printf("precalc done.\n");
	atmos_set_intensity_profile(0);
 	intensity_table = atmos_get_intensity_profile();
	render_set_cam_angle(0.);
	render_set_cam_tilt(0.);
	return 0;
}
