#include <SDL/SDL.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "render.h"

#define MOTION_BLUR 1

//struct pixel_spectrum camera_screen[Y_RES*X_RES];
float rgbuf[Y_RES*X_RES*3];
unsigned char crgbuf[Y_RES*X_RES*3], crgbuf_new[Y_RES*X_RES*3];

int write_ppm(unsigned char *crgbuf, char *name) 
{
	int ret = -1;
	FILE *ppm;

	/* write the RGB buffer to Portable PixMap file.. */
	ppm = fopen(name, "wb");
	if (ppm==NULL) {
		printf("failed to create ppm file!\n");
		goto error;
	}
	fprintf(ppm, "P6\n");
	fprintf(ppm, "# generated by the aurora simulator by Pieter van der Meer.\n");
	fprintf(ppm, "%d %d\n",X_RES,Y_RES);
	fprintf(ppm, "255\n");
	int n_written = fwrite(crgbuf, 1, X_RES*Y_RES*3, ppm);
	if (n_written < X_RES*Y_RES) {
		printf("failed to write to ppm file!\n");
		goto error;
	}
	ret = 0;

error:
	if (ppm)
		fclose(ppm);
	return ret;
}

/*----------------------------------------------------------------------------*/
/* main test routine                                                          */
/*----------------------------------------------------------------------------*/

int main(int argc, char *argv)
{
	char filename[64];
	int frame_nr;

	memset(crgbuf_new, 0, X_RES*Y_RES*3*sizeof(unsigned char));
	if (render_init()) {
		printf("render_init failed!\n");
		goto end;
	}

	for (frame_nr=0; frame_nr<40; frame_nr++) {

		printf("frame=[%04d]\n", frame_nr);

		memset(rgbuf, 0, X_RES*Y_RES*3*sizeof(float));

		#if MOTION_BLUR
		int i;
		for (i=0; i<X_RES*Y_RES*3; i++) {
			rgbuf[i] *= .9;
		}
		render(crgbuf, rgbuf, (float) frame_nr * .001, 1, .1);
		#else
		memset(rgbuf, 0, X_RES*Y_RES*3*sizeof(float));
		render(crgbuf, rgbuf, (float) frame_nr * .001, 1, 1.);
		#endif

		sprintf(filename, "aurora%03d.ppm", frame_nr);
		if (write_ppm(crgbuf, filename)) {
			printf("failed to write '%s'\n", filename);
			return -1;
		}

	}

end:	return 0;
}
