#include <SDL/SDL.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "atmos.h"
#include "render.h"

#define MOTION_BLUR 1

#define SDL_WIDTH X_RES
#define SDL_HEIGHT Y_RES

SDL_Surface *screen;

//struct pixel_spectrum camera_screen[Y_RES*X_RES];
float rgbuf[Y_RES*X_RES*3];
unsigned char crgbuf[Y_RES*X_RES*3], crgbuf_new[Y_RES*X_RES*3];

/*----------------------------------------------------------------------------*/
/* main test routine                                                          */
/*----------------------------------------------------------------------------*/

void Slock(SDL_Surface *screen)
{
	if ( SDL_MUSTLOCK(screen) ) {
		if ( SDL_LockSurface(screen) < 0 ) {
			return;
		}
	}
}

void Sulock(SDL_Surface *screen)
{
	if ( SDL_MUSTLOCK(screen) ) {
		SDL_UnlockSurface(screen);
	}
}

SDL_Surface *font;

void draw_text(int x, int y, const char* text, SDL_Surface *screen)
{
    SDL_Rect src;
    SDL_Rect dest;
    while (*text) {
        src.x = ((*text) & 0xF) * 6;
        src.y = (((*text) >> 4) - 2) * 8;
        src.w = 6;
        src.h = 8;
        dest.x = x;
        dest.y = y;
        SDL_BlitSurface(font, &src, screen, &dest);
        text++;
        x += 6;
    }
}

int blow, gas, v_dir, u_dir, legend_on=1;

int get_from_ui()
{
	int i,j,x,y;

	SDL_Event event;
	while ( SDL_PollEvent(&event) )	{
		if (event.type == SDL_QUIT) {
			return -1;
		}
		switch(event.type) {
			case SDL_KEYDOWN:
			switch(event.key.keysym.sym) {
				case SDLK_ESCAPE:
				return -1;
				case SDLK_SPACE:
				legend_on^=1;
				break;
				case SDLK_RETURN:
				gas=1;
				break;
				case SDLK_LEFT:
				v_dir=-1;
				printf("left\n");
				break;
				case SDLK_RIGHT:
				v_dir=+1;
				printf("right\n");
				break;
				case SDLK_UP:
				u_dir=-1;
				printf("up\n");
				break;
				case SDLK_DOWN:
				u_dir=+1;
				printf("down\n");
				break;
				case SDLK_F1:
				atmos_set_intensity_profile(0);
				break;
				case SDLK_F2:
				atmos_set_intensity_profile(1);
				break;
				case SDLK_F3:
				atmos_set_intensity_profile(2);
				break;
				case SDLK_F4:
				atmos_set_intensity_profile(3);
				break;
			}
			break;
			case SDL_KEYUP:
			switch(event.key.keysym.sym) {
				case SDLK_RETURN:
				gas=0;
				break;
				case SDLK_LEFT:
				v_dir=0;
				break;
				case SDLK_RIGHT:
				v_dir=0;
				break;
				case SDLK_UP:
				u_dir=0;
				break;
				case SDLK_DOWN:
				u_dir=0;
				break;
			}
			break;
			// other event types
		}
	}

	return 0;
}

int main(int argc, char *argv)
{
	char filename[64];
	int frame_nr, old_fr_count;
	int simulating=1;
	int ret=-1;
	long old_ticks=0, last_msg_ticks=0;
	float cam_angle = 0., cam_tilt = 0.;

	/* SDL initialisation */
	atexit(SDL_Quit);

	if( SDL_Init(SDL_INIT_JOYSTICK|SDL_INIT_VIDEO /*|SDL_INIT_AUDIO */) <0 ) {
		printf("Unable to init SDL: %s\n", SDL_GetError());
		goto end;
	}
	SDL_JoystickOpen(0);
	SDL_ShowCursor(0);

	/* TODO: SDL_HWSURFACE also possible? */
	screen = SDL_SetVideoMode(SDL_WIDTH, SDL_HEIGHT, 24, SDL_SWSURFACE /*|SDL_DOUBLEBUF */);
	if ( screen == NULL ) {
		printf("Unable to set %dX%d video: %s\n", SDL_WIDTH, SDL_HEIGHT, SDL_GetError());
		goto end;
	}

	SDL_Surface *temp = SDL_LoadBMP("TinyFont.bmp");
	if (temp == NULL) {
		printf("Unable to load bitmap: %s\n", SDL_GetError());
		goto end;
	}
	font = SDL_DisplayFormat(temp);
	/* clear screen*/
	SDL_FillRect(screen, NULL, 0);

	if (render_init()) {
		printf("render_init failed!\n");
		goto end;
	}

	char fps_msg[32] = "";
	for (frame_nr=0; simulating; frame_nr++) {

		printf("[%05d]\n", frame_nr);

		Slock(screen);
	
		if (get_from_ui()) {
			simulating=0;
		}
		cam_angle += (float) v_dir * .05;
		cam_tilt += (float) u_dir * .05;
		if (cam_tilt < -.5) {
			cam_tilt=-.5;
		} else if (cam_tilt > 1.) {
			cam_tilt=+1.;
		}
		cam_angle = fmod(cam_angle, 2.*M_PI);
		render_set_cam_angle(cam_angle);
		render_set_cam_tilt(cam_tilt);

		Uint8 *pixels;
		int pitch;
		pixels = screen->pixels;
		pitch = screen->pitch;
		long ticks = SDL_GetTicks();

		#if MOTION_BLUR
		int i;
		for (i=0; i<X_RES*Y_RES*3; i++) {
			rgbuf[i] *= .9;
		}
		render(pixels, rgbuf, (float) ticks * .0001, 1, .1);
		#else
		memset(rgbuf, 0, X_RES*Y_RES*3*sizeof(float));
		render(pixels, rgbuf, (float) ticks * .0001, 1, 1.);
		#endif

		if (old_ticks && (ticks-last_msg_ticks>5000)) {
			//int t = ticks - old_ticks;
			//sprintf(fps_msg, "%d fps  ", 1000/t);
			int t = ticks - last_msg_ticks;
			int f = frame_nr - old_fr_count;
			sprintf(fps_msg, "%03d fps", (f*1000)/t);
			last_msg_ticks=ticks;
			old_fr_count=frame_nr;
		}
		old_ticks = ticks;

		/* draw legend and usage */
		if (legend_on) {
			draw_text(0, 0, "arrows to change curtain angle, function keys to change form, [esc] to stop", screen);
			char cam_str[32];
			sprintf(cam_str, "[%f,%f]", cam_angle, cam_tilt);
			draw_text(0, 10, cam_str, screen);
			draw_text(0, 20, fps_msg, screen);
		}

		/* SDL display shit */
		Sulock(screen);
		SDL_Flip(screen);
	}
	ret=0;

end:	SDL_Quit();  // Shut down the SDL library, freeing resources.
	return ret;
}
