#include <math.h>
#include <stdio.h>
#include "transform.h"

/*----------------------------------------------------------------------------*/
/* visualisation: 3D space -> 2D camera space transform                       */
/*----------------------------------------------------------------------------*/

void create_matrix(float *mat, float a, float b, float c)
{
	mat[0*N_DIM+0] = + cos(b)*cos(c);
	mat[0*N_DIM+1] = - cos(b)*sin(c);
	mat[0*N_DIM+2] = + sin(b);
	mat[1*N_DIM+0] = + sin(a)*sin(b)*cos(c)+cos(a)*sin(c);
	mat[1*N_DIM+1] = + cos(a)*cos(c)-sin(a)*sin(b)*sin(c);
	mat[1*N_DIM+2] = - sin(a)*cos(b);
	mat[2*N_DIM+0] = + sin(a)*sin(c)-cos(a)*sin(b)*cos(c);
	mat[2*N_DIM+1] = + cos(a)*sin(b)*sin(c)+sin(a)*cos(c);
	mat[2*N_DIM+2] = + cos(a)*cos(b);
}

void print_matrix(float *mat)
{
	printf("[%f %f %f]\n", mat[0*N_DIM+0], mat[0*N_DIM+1], mat[0*N_DIM+2]);
	printf("[%f %f %f]\n", mat[1*N_DIM+0], mat[1*N_DIM+1], mat[1*N_DIM+2]);
	printf("[%f %f %f]\n", mat[2*N_DIM+0], mat[2*N_DIM+1], mat[2*N_DIM+2]);
}

#define CENTERED 1

void transform_points(float *dst, float *src, int n, float *mat, int *center, float dist, float *centre)
{

	float *d2 = dst;
	float *s2 = src;
	int i;

	/* 3D transformation */
	#if CENTERED
	float trs[N_DIM]; 
	for (i=0; i<n; i++) {
		int j, k;
		for (k=0; k<N_DIM; k++)
			trs[k] = src[k] - centre[k];
		for (j=0; j<N_DIM; j++) {
			float prod = 0;
			for (k=0; k<N_DIM; k++)
				prod += mat[j*N_DIM+k] * trs[k];
			*dst++ = prod;
		}
		src+=N_DIM;
		for (k=0; k<N_DIM; k++)
			dst[-1-k] = dst[-1-k] + centre[N_DIM-1-k];
	}
	#else
	for (i=0; i<n; i++) {
		int j;
		for (j=0; j<N_DIM; j++) {
			float prod = 0;
			int k;
			for (k=0; k<N_DIM; k++)
				prod += mat[j*N_DIM+k] * src[k];
			*dst++ = prod;
		}
		src+=N_DIM;
	}
	#endif
	src=s2;
	dst=d2;

	/*
         * project onto camera plane: stupid math conventions: z-axis is up.. pah..
         */
	for (i=0; i<n; i++) {
		float x = dst[0];
		float z = dst[1];
		float y = dst[2];
		x *= dist / (z + dist);
		y *= dist / (z + dist);
		x += (float) center[0];
		y += (float) center[1];
		dst[0] = x;
		dst[1] = y;
		dst[2] = z;
		dst += N_DIM;
	}
}
